<?php

namespace App\Models;

use Exception;
use Illuminate\Database\Eloquent\Collection as DatabaseCollection;
use Illuminate\Database\Eloquent\Model;
use Illuminate\Database\Eloquent\Relations\BelongsTo;
use Illuminate\Database\Eloquent\Relations\HasMany;
use Illuminate\Database\Eloquent\Relations\HasOne;
use Illuminate\Database\Eloquent\Relations\MorphTo;
use Illuminate\Support\Arr;
use Illuminate\Support\Collection;

use App\Traits\NextKey;
use Leenooks\Carbon;

class Service extends Model
{
	use NextKey;
	const RECORD_ID = 'service';
	public $incrementing = FALSE;

	const CREATED_AT = 'date_orig';
	const UPDATED_AT = 'date_last';

	protected $dates = [
		'date_last_invoice',
		'date_next_invoice'.
		'date_start',
		'date_end',
	];

	public $dateFormat = 'U';
	protected $table = 'ab_service';

	protected $casts = [
		'order_info'=>'array',
	];

	protected $appends = [
		'account_name',
		'admin_service_id_url',
		'billing_price',
		'name_short',
		'next_invoice',
		'product_category',
		'product_name',
		'service_id',
		'service_id_url',
		'status',
	];

	protected $visible = [
		'account_name',
		'admin_service_id_url',
		'active',
		'billing_price',
		'data_orig',
		'id',
		'name_short',
		'next_invoice',
		'product_category',
		'product_name',
		'service_id',
		'service_id_url',
		'status',
	];

	private $inactive_status = [
		'CANCELLED',
		'ORDER-REJECTED',
		'ORDER-CANCELLED',
	];

	/**
	 * Valid status shows the applicable next status for an action on a service
	 * Each status can be
	 * + Approved, to proceed to the next valid status'
	 * + Held, to a holding pattern status
	 * + Rejected, reverted to an different status
	 * + Cancel, to progress down a decomission route
	 * + Updated, stay on the current status with new information
	 *
	 * @var array
	 */
	private $valid_status = [
		// Order Submitted
		'ORDER-SUBMIT' => ['approve'=>'ORDER-SENT','hold'=>'ORDER-HOLD','reject'=>'ORDER-REJECTED','cancel'=>'ORDER-CANCELLED'],
		// Order On Hold (Reason)
		'ORDER-HOLD' => ['release'=>'ORDER-SUBMIT','update_reference'=>'ORDER-SENT'],
		// Order Rejected (Reason)
		'ORDER-REJECTED' => [],
		// Order Cancelled
		'ORDER-CANCELLED' => [],
		// Order Sent to Supplier
		'ORDER-SENT' => ['update_reference'=>'ORDER-SENT','confirm'=>'ORDERED'],
		// Order Confirmed by Supplier
		'ORDERED' => ['update_reference'=>'ORDER-SENT'],
	];

	/**
	 * Account the service belongs to
	 *
	 * @return BelongsTo
	 */
	public function account()
	{
		return $this->belongsTo(Account::class);
	}

	/**
	 * Return automatic billing details
	 *
	 * @return HasOne
	 */
	public function billing()
	{
		return $this->hasOne(AccountBilling::class);
	}

	/**
	 * Return Charges associated with this Service
	 *
	 * @return HasMany
	 */
	public function charges()
	{
		return $this->hasMany(Charge::class)
			->where('active','=',TRUE)
			->orderBy('date_orig');
	}

	// @todo changed to invoiced_items
	public function invoice_items($active=TRUE)
	{
		$query = $this->hasMany(InvoiceItem::class)
			->where('item_type','=',0)
			->orderBy('date_orig');

		if ($active)
			$query->where('active','=',TRUE);

		return $query;
	}

	/**
	 * Invoices for this service
	 *
	 */
	public function invoices($active=TRUE)
	{
		$query = $this->hasManyThrough(Invoice::class,InvoiceItem::class,NULL,'id',NULL,'invoice_id')
			->distinct('id')
			->where('ab_invoice.site_id','=',$this->site_id)
			->where('ab_invoice_item.site_id','=',$this->site_id)
			->orderBy('date_orig')
			->orderBy('due_date');

		if ($active)
			$query->where('ab_invoice_item.active','=',TRUE)
				->where('ab_invoice.active','=',TRUE);

		return $query;
	}

	/**
	 * Account that ordered the service
	 *
	 * @return BelongsTo
	 */
	public function orderedby()
	{
		return $this->belongsTo(Account::class);
	}

	/**
	 * Product of the service
	 *
	 * @return BelongsTo
	 */
	public function product()
	{
		return $this->belongsTo(Product::class);
	}

	/**
	 * Return a child model with details of the service
	 *
	 * @return MorphTo
	 */
	public function type()
	{
		if (! $this->model)
			abort(500,'Missing Model in',['service'=>$this]);

		return $this->morphTo(null,'model','id','service_id');
	}

	/** SCOPES **/

	/**
	 * Only query active categories
	 */
	public function scopeActive($query)
	{
		return $query->where(function () use ($query) {
			$query->where('active',TRUE)
				->orWhereNotIn('order_status',$this->inactive_status);
		});
	}

	/**
	 * Find inactive services.
	 *
	 * @param $query
	 * @return mixed
	 */
	public function scopeInActive($query)
	{
		return $query->where(function () use ($query) {
			$query->where('active',FALSE)
				->orWhereIn('order_status',$this->inactive_status);
		});
	}

	/**
	 * Enable to perform queries without eager loading
	 *
	 * @param $query
	 * @return mixed
	 */
	public function scopeNoEagerLoads($query){
		return $query->setEagerLoads([]);
	}

	/**
	 * Search for a record
	 *
	 * @param        $query
	 * @param string $term
	 * @return
	 */
	public function scopeSearch($query,string $term)
	{
		return $query->where('id','like','%'.$term.'%');
	}

	/** ATTRIBUTES **/

	/**
	 * Name of the account for this service
	 *
	 * @return mixed
	 */
	public function getAccountNameAttribute(): string
	{
		return $this->account->name;
	}

	/**
	 * @deprecated Use getUrlAdminAttribute()
	 */
	public function getAdminServiceIdUrlAttribute()
	{
		return $this->getUrlAdminAttribute();
	}

	/**
	 * Return the auto billing details
	 *
	 * @return mixed
	 */
	public function getAutoPayAttribute()
	{
		return $this->billing;
	}

	public function getBillingPriceAttribute(): float
	{
		// @todo Temporary for services that dont have recur_schedule set.
		if (is_null($this->recur_schedule) OR is_null($this->product->price($this->recur_schedule)))
			$this->price=0;

		return $this->addTax(is_null($this->price) ? $this->product->price($this->recur_schedule) : $this->price);
	}

	/**
	 * Return the service billing period
	 *
	 * @return string
	 */
	public function getBillingPeriodAttribute(): string
	{
		return Arr::get($this->product->PricePeriods(),$this->recur_schedule,'Unknown');
	}

	/**
	 * Date the service expires, also represents when it is paid up to
	 *
	 * @return string
	 */
	public function getExpiresAttribute(): string
	{
		return 'TBA';
	}

	/**
	 * Return the date for the next invoice
	 *
	 * @todo This function negates the need for date_next_invoice
	 * @return Carbon|string
	 */
	public function getInvoiceNextAttribute()
	{
		$last = $this->getInvoiceToAttribute();
		$date = $last ? $last->addDay() : Carbon::now();

		return request()->wantsJson() ? $date->format('Y-m-d') : $date;
	}

	/**
	 * Return the end date for the next invoice
	 *
	 * @return mixed
	 * @throws Exception
	 */
	public function getInvoiceNextEndAttribute()
	{
		switch ($this->recur_schedule) {
			// Weekly
			case 0: $date = $this->product->price_recurr_strict
				? $this->getInvoiceNextAttribute()->endOfWeek()
				: $this->getInvoiceNextAttribute()->addWeek()->subDay();
				break;

			// Monthly
			case 1:
				$date = $this->product->price_recurr_strict
					? $this->getInvoiceNextAttribute()->endOfMonth()
					: $this->getInvoiceNextAttribute()->addMonth()->subDay();
				break;

			// Quarterly
			case 2:
				$date = $this->product->price_recurr_strict
					? $this->getInvoiceNextAttribute()->endOfQuarter()
					: $this->getInvoiceNextAttribute()->addQuarter()->subDay();
				break;

			// Half Yearly
			case 3:
				$date = $this->product->price_recurr_strict
					? $this->getInvoiceNextAttribute()->endOfHalf()
					: $this->getInvoiceNextAttribute()->addQuarter(2)->subDay();
				break;

			// Yearly
			case 4:
				$date = $this->product->price_recurr_strict
					? $this->getInvoiceNextAttribute()->endOfYear()
					: $this->getInvoiceNextAttribute()->addYear()->subDay();
				break;

			// Two Yearly
			// NOTE: price_recurr_strict ignored
			case 5: $date = $this->getInvoiceNextAttribute()->addYear(2)->subDay(); break;

			// Three Yearly
			// NOTE: price_recurr_strict ignored
			case 6: $date = $this->getInvoiceNextAttribute()->addYear(3)->subDay(); break;

			default: throw new Exception('Unknown recur_schedule');
		}

		return $date;
	}

	public function getInvoiceNextQuantityAttribute()
	{
		// If we are not rounding to the first day of the cycle, then it is always a full cycle
		if (! $this->product->price_recurr_strict)
			return 1;

		$n = $this->invoice_next->diff($this->invoice_next_end)->days+1;

		switch ($this->recur_schedule) {
			// Weekly
			case 0:
				$d = $this->invoice_next_end->diff($this->invoice_next_end->startOfWeek())->days;
				break;

			// Monthly
			case 1:
				$d = $this->invoice_next_end->diff($this->invoice_next_end->startOfMonth())->days;
				break;

			// Quarterly
			case 2:
				$d = $this->invoice_next_end->diff($this->invoice_next_end->startOfQuarter())->days;
				break;

			// Half Yearly
			case 3:
				$d = $this->invoice_next_end->diff($this->invoice_next_end->startOfHalf())->days;
				break;

			// Yearly
			case 4:
				$d = $this->invoice_next_end->diff($this->invoice_next_end->startOfYear())->days;
				break;

			// Two Yearly
			case 5:
				$d = $this->invoice_next_end->diff($this->invoice_next_end->subyear(2))->days-1;
				break;

			// Three Yearly
			case 6:
				$d = $this->invoice_next_end->diff($this->invoice_next_end->subyear(3))->days-1;
				break;

			default: throw new Exception('Unknown recur_schedule');
		}

		// Include the start date and end date.
		$d += 1;

		return round($n/$d,2);
	}

	/**
	 * Get the date that the service has been invoiced to
	 */
	public function getInvoiceToAttribute()
	{
		return $this->invoice_items->count() ? $this->invoice_items->last()->date_stop : NULL;
	}

	public function getNameAttribute(): string
	{
		return $this->product->name_short.': '.$this->getNameShortAttribute();
	}

	/**
	 * Return the short name for the service.
	 *
	 * EG:
	 *  For ADSL, this would be the phone number,
	 *  For Hosting, this would be the domain name, etc
	 * @deprecated
	 */
	public function getNameShortAttribute()
	{
		return $this->model ? $this->type->name : 'NAME UNKNOWN';
	}

	/**
	 * @deprecated see getInvoiceNextAttribute()
	 */
	public function getNextInvoiceAttribute()
	{
		return $this->getInvoiceNextAttribute();
	}

	/**
	 * This function will present the Order Info Details
	 */
	public function getOrderInfoDetailsAttribute(): string
	{
		if (! $this->order_info)
			return '';

		$result = '';

		foreach ($this->order_info as $k=>$v)
		{
			if (in_array($k,['order_reference']))
				continue;

			$result .= sprintf('%s: <b>%s</b><br>',ucfirst($k),$v);
		}

		return $result;
	}

	/**
	 * Work out when this service has been paid to.
	 *
	 * @todo This might need to be optimised
	 */
	public function getPaidToAttribute()
	{
		foreach ($this->invoices->reverse() as $o) {
			if ($o->due == 0) {
				return $o->items
					->filter(function($item) {
						return $item->item_type === 0;
					})
					->last()
					->date_stop;
			}
		}
	}

	/**
	 * Get the Product's Category for this service
	 *
	 */
	public function getProductCategoryAttribute(): string
	{
		return $this->product->category;
	}

	/**
	 * Get the Product's Short Name for the service
	 *
	 * @return string
	 */
	public function getProductNameAttribute(): string
	{
		return $this->product->name($this->account->language);
	}

	public function getRecurScheduleAttribute($value): int
	{
		// If recur_schedule not set, default to 2
		return $value ?? 2;
	}

	/**
	 * @deprecated see getSIDAttribute()
	 */
	public function getServiceIdAttribute(): string
	{
		return $this->getSIDAttribute();
	}

	/**
	 * @deprecated see getUrlUserAttribute()
	 */
	public function getServiceIdUrlAttribute()
	{
		return $this->getUrlUserAttribute();
	}

	/**
	 * @deprecated see getServiceIdAttribute()
	 */
	public function getServiceNumberAttribute()
	{
		return $this->getSIDAttribute();
	}

	/**
	 * Services Unique Identifier
	 *
	 * @return string
	 */
	public function getSIDAttribute(): string
	{
		return sprintf('%02s-%04s.%05s',$this->site_id,$this->account_id,$this->id);
	}

	/**
	 * Return the service description.
	 * For:
	 *  + Broadband, this is the service address
	 *  + Domains, blank
	 *  + Hosting, blank
	 *  + SSL, blank
	 *
	 * @return string
	 */
	public function getSDescAttribute(): string
	{
		return $this->type->service_description ?: 'Service Description NOT Defined for :'.$this->type->type;
	}

	/**
	 * Return the service name.
	 * For:
	 *  + Broadband, this is the service number
	 *  + Domains, this is the full domain name
	 *  + Hosting, this is the full domain name
	 *  + SSL, this is the DN
	 *
	 * @return string
	 */
	public function getSNameAttribute(): string
	{
		return $this->type->service_name ?: 'Service Name NOT Defined for :'.$this->type->type;
	}

	/**
	 * Return the service product type
	 * This is used for view specific details
	 *
	 * @return string
	 */
	public function getSTypeAttribute(): string
	{
		switch($this->product->model) {
			case 'App\Models\Product\Adsl': return 'broadband';
			default: return $this->type->type;
		}
	}

	/**
	 * Return the Service Status
	 *
	 * @return string
	 */
	public function getStatusAttribute(): string
	{
		if (! $this->order_status)
			return $this->active ? 'ACTIVE' : 'INACTIVE';
		else
			return $this->order_status;
	}

	/**
	 * Return the detailed order Status, with order reference numbers.
	 *
	 * @return string
	 */
	public function getStatusDetailAttribute(): string
	{
		return in_array($this->order_status,['ORDER-SENT','ORDER-HOLD','ORDERED'])
			? sprintf('%s: <span class="white-space: nowrap"><small><b>#%s</b></small></span>',$this->order_status,Arr::get($this->order_info,'order_reference','Unknown'))
			: '';
	}

	/**
	 * Return a HTML status box
	 *
	 * @return string
	 */
	public function getStatusHTMLAttribute(): string
	{
		$class = NULL;

		if ($this->isPending())
			$class = 'badge-warning';

		else
			switch ($this->status)
			{
				case 'ACTIVE':
					$class = 'badge-success';
					break;
				case 'INACTIVE':
					$class = 'badge-danger';
					break;
			}

		return $class
			? sprintf('<span class="badge %s">%s</span>',$class,$this->status)
			: $this->status;
	}

	/**
	 * URL used by an admin to administer the record
	 *
	 * @return string
	 */
	public function getUrlAdminAttribute(): string
	{
		return sprintf('<a href="/a/service/%s">%s</a>',$this->id,$this->service_id);
	}

	/**
	 * URL used by an user to see the record
	 *
	 * @return string
	 */
	public function getUrlUserAttribute(): string
	{
		return sprintf('<a href="/u/service/%s">%s</a>',$this->id,$this->service_id);
	}

	/** SETTERS **/

	public function setDateOrigAttribute($value)
	{
		$this->attributes['date_orig'] = $value->timestamp;
	}

	public function setDateLastAttribute($value)
	{
		$this->attributes['date_last'] = $value->timestamp;
	}

	/** FUNCTIONS **/

	/**
	 * Add applicable tax to the cost
	 *
	 * @todo This needs to be calculated, not fixed at 1.1
	 * @param float $value
	 * @return float
	 */
	private function addTax(float $value): float
	{
		return round($value*1.1,2);
	}

	public function invoices_due(): DatabaseCollection
	{
		$this->load('invoice_items.invoice');

		return $this->invoice_items->filter(function($item) {
			return $item->invoice->due > 0;
		});
	}

	/**
	 * Determine if a service is active. It is active, if active=1, or the order_status is not in inactive_status[]
	 *
	 * @return bool
	 * @todo Remove active and have order_status reflect whether active or not
	 */
	public function isActive(): bool
	{
		return $this->active OR ($this->order_status AND ! in_array($this->order_status,$this->inactive_status));
	}

	/**
	 * Should this service be invoiced soon
	 *
	 * @todo get the number of days from account setup
	 * @return bool
	 */
	public function isInvoiceDueSoon($days=30): bool
	{
		return (! $this->external_billing) AND (! $this->suspend_billing) AND $this->getInvoiceNextAttribute()->lessThan(now()->addDays($days));
	}

	/**
	 * Identify if a service is being ordered
	 *
	 * @return bool
	 */
	public function isPending(): bool
	{
		return ! $this->active
			AND ! is_null($this->order_status)
			AND ! in_array($this->order_status,array_merge($this->inactive_status,['INACTIVE']));
	}

	/**
	 * Generate a collection of invoice_item objects that will be billed for the next invoice
	 *
	 * @return Collection
	 * @throws Exception
	 */
	public function next_invoice_items(): Collection
	{
		$result = collect();

		$o = new InvoiceItem;

		// If the service is active, there will be service charges
		if ($this->active or $this->isPending()) {
			$o->active = TRUE;
			$o->service_id = $this->id;
			$o->product_id = $this->product_id;
			$o->item_type = 0;
			$o->price_base = $this->price ?: $this->product->price($this->recur_schedule); // @todo change to a method in this class
			$o->recurring_schedule = $this->recur_schedule;
			$o->date_start = $this->invoice_next;
			$o->date_stop = $this->invoice_next_end;
			$o->quantity = $this->invoice_next_quantity;

			$o->addTaxes();
			$result->push($o);
		}

		// If pending, add any connection charges
		if ($this->isPending()) {
			$o = new InvoiceItem;
			$o->active = TRUE;
			$o->service_id = $this->id;
			$o->product_id = $this->product_id;
			$o->item_type = 4;
			$o->price_base = $this->price ?: $this->product->price($this->recur_schedule,'price_setup'); // @todo change to a method in this class
			//$o->recurring_schedule = $this->recur_schedule;
			$o->date_start = $this->invoice_next;
			$o->date_stop = $this->invoice_next;
			$o->quantity = 1;

			$o->addTaxes();
			$result->push($o);
		}

		//  Add additional charges
		foreach ($this->charges->filter(function($item) { return ! $item->processed; }) as $oo) {
			$o = new InvoiceItem;
			$o->active = TRUE;
			$o->service_id = $oo->service_id;
			$o->product_id = $this->product_id;
			$o->quantity = $oo->quantity;
			$o->item_type = $oo->type;
			$o->price_base = $oo->amount;
			$o->date_start = $oo->date_charge;
			$o->date_stop = $oo->date_charge;
			$o->module_id = 30; // @todo This shouldnt be hard coded
			$o->module_ref = $oo->id;

			$o->addTaxes();
			$result->push($o);
		}

		return $result;
	}

	/**
	 * @todo
	 * @param string $status
	 * @return $this
	 */
	public function nextStatus(string $status) {
		if ($x=$this->validStatus($status))
		{
			$this->order_status = $x;
			$this->save();

			return $this;
		}

		abort(500,'Next Status not set up for:'.$this->order_status);
	}

	/**
	 * This function will return the associated service model for the product type
	 * @deprecated use $this->type
	 */
	private function ServicePlugin()
	{
		// @todo: All services should be linked to a product. This might require data cleaning for old services not linked to a product.
		if (! is_object($this->product))
			return NULL;

		switch ($this->product->prod_plugin_file)
		{
			case 'ADSL': return $this->service_adsl;
			case 'DOMAIN': return $this->service_domain;
			case 'HOST': return $this->service_host;
			case 'SSL': return $this->service_ssl;
			case 'VOIP': return $this->service_voip;

			default: return NULL;
		}
	}

	/**
	 * Return if the proposed status is valid.
	 *
	 * @param string $status
	 * @return string | NULL
	 */
	private function testNextStatusValid(string $status)
	{
		return Arr::get(Arr::get($this->valid_status,$this->order_status,[]),$status,NULL);
	}

	/**
	 * @deprecated use testNextStatusValid()
	 */
	public function validStatus(string $status)
	{
		return $this->testNextStatusValid($status);
	}
}